/** Record<string, DataType> alias */
type DataRecord = {
    [P in string]: DataType;
};
/** Defining posible data type rather than using unknown */
type DataType = DataRecord | DataRecord[] | DataType[] | string | number | boolean | null | undefined;
type OmitUtilityProps<K> = K extends string ? K extends `${'$'}${infer _X}` | 'toJSON' ? never : K : K;

type parseOptions<T> = {
    default?: T;
    parseHandler?: (value: DataType) => T;
};
type InferArrayOptions = {
    /** when true on inferEach or inferAt error will be ignored and will only return successfully inferred row only. (warn messages will still console logged) */
    skipError?: boolean;
};

/**
 * Inferer: safely infering data for managing and transforming data
 */
declare class Inferer {
    _data: DataType;
    constructor(data: DataType);
    /**
     * create error object
     *
     * @private
     * @param {...string[]} tags
     * @return {Error}
     * @memberof Inferer
     */
    private _createError;
    /**
     * fallback handler for each infer type function
     *
     * @private
     * @template T
     * @param {string} path
     * @param {string} type
     * @param {DataType} value
     * @param {parseOptions<T>} options
     * @return {T}
     * @memberof Inferer
     */
    private _fallbackHandler;
    /**
     * Get raw value based on object path
     */
    get(path?: string): DataType;
    /**
     * check given path if value exist or not.
     *
     * @param {string} path
     * @param {(data: DataType) => boolean} [validator=(v) => v !== undefined] default to `(v) => v !== undefined`
     * @return {boolean}
     */
    has(path: string, validator?: (data: DataType) => boolean): boolean;
    /**
     * Infer object path in runtime either value is string or throw an error
     *
     * @param {string} [path='']
     * @param {parseOptions<string>} [options={}]
     * @return {string}
     * @memberof Inferer
     */
    string(path?: string, options?: parseOptions<string>): string;
    /**
     * alias for string inferer
     */
    str(path?: string, options?: parseOptions<string>): string;
    /**
     * Infer object path in runtime either value is number (int|float) or throw an error
     *
     * @param {string} [path='']
     * @param {(parseOptions<number> & { float?: boolean })} [options={}]
     * @return {number}
     * @memberof Inferer
     */
    number(path?: string, options?: parseOptions<number> & {
        float?: boolean;
    }): number;
    /**
     * alias for number inferer
     */
    num(path?: string, options?: parseOptions<number> & {
        float?: boolean;
    }): number;
    /**
     * Infer object path in runtime either value is boolean or throw an error
     *
     * @param {string} [path='']
     * @param {parseOptions<boolean>} [options={}]
     * @return {boolean}
     * @memberof Inferer
     */
    boolean(path?: string, options?: parseOptions<boolean> & {
        force?: boolean;
    }): boolean;
    /**
     * alias for boolean inferer
     */
    bool(path?: string, options?: parseOptions<boolean>): boolean;
    /**
     * Infer object path in runtime accepting value of string or number then return as a string.
     *
     * Return error when neither string or number found.
     * Usefull for id field that you are unsure is it number or string value.
     *
     * @param {string} [path='']
     * @param {parseOptions<string>} [options={}]
     * @return {string}
     * @memberof Inferer
     */
    stringNumber(path?: string, options?: parseOptions<string>): string;
    /** alias for stringOrNumber */
    snum(path?: string, options?: parseOptions<string>): string;
    /**
     * Infer object path in runtime is array then return as InfererArray object.
     *
     * @param {boolean} [defaultEmpty=false] when true return will fallback to empty array. On false will throw error. default: false.
     */
    array(path?: string, options?: parseOptions<InfererArray> & {
        defaultEmpty?: boolean;
    } & InferArrayOptions): InfererArray;
    /**
     * Infer object path when path has value it will parse or transform with given function or throw an error when value is null or undefined.
     */
    parse<T = DataType>(path: string, fn: (data: unknown) => T, options?: parseOptions<T>): T;
    /**
     * Infer object path and fallback to given preset value and type if type not match
     */
    maybe<T>(path: string, preset: T): {
        /** check if given path is a string or return preset value */
        string: (options?: parseOptions<string>) => string | T;
        /** check if given path is a string or return preset value */
        str: (options?: parseOptions<string>) => string | T;
        /** check if given path is a number or return preset value */
        number: (options?: parseOptions<number>) => number | T;
        /** check if given path is a number or return preset value */
        num: (options?: parseOptions<number>) => number | T;
        /** check if given path is a boolean or return preset value */
        boolean: (options?: parseOptions<boolean> & {
            force?: boolean;
        }) => boolean | T;
        /** check if given path is a boolean or return preset value */
        bool: (options?: parseOptions<boolean> & {
            force?: boolean;
        }) => boolean | T;
        /** check if given path is a string, number or return preset value */
        stringNumber: (options?: parseOptions<string>) => string | T;
        /** check if given path is a string, number or return preset value */
        snum: (options?: parseOptions<string>) => string | T;
        /** check if given path has value then parse using given function or return preset value */
        parse: <V>(fn: (data: unknown) => V, options?: parseOptions<V>) => T | V;
        /** check if given path is array then return InfererArray or return preset value */
        array: (options?: parseOptions<InfererArray> & {
            defaultEmpty?: boolean;
        }) => InfererArray | T;
    };
    /**
     * alias or shortcut for maybe with preset value null
     */
    nullable(path?: string): {
        /** check if given path is a string or return preset value */
        string: (options?: parseOptions<string>) => string | null;
        /** check if given path is a string or return preset value */
        str: (options?: parseOptions<string>) => string | null;
        /** check if given path is a number or return preset value */
        number: (options?: parseOptions<number>) => number | null;
        /** check if given path is a number or return preset value */
        num: (options?: parseOptions<number>) => number | null;
        /** check if given path is a boolean or return preset value */
        boolean: (options?: parseOptions<boolean> & {
            force?: boolean | undefined;
        }) => boolean | null;
        /** check if given path is a boolean or return preset value */
        bool: (options?: parseOptions<boolean> & {
            force?: boolean | undefined;
        }) => boolean | null;
        /** check if given path is a string, number or return preset value */
        stringNumber: (options?: parseOptions<string>) => string | null;
        /** check if given path is a string, number or return preset value */
        snum: (options?: parseOptions<string>) => string | null;
        /** check if given path has value then parse using given function or return preset value */
        parse: <V>(fn: (data: unknown) => V, options?: parseOptions<V>) => V | null;
        /** check if given path is array then return InfererArray or return preset value */
        array: (options?: parseOptions<InfererArray> & {
            defaultEmpty?: boolean | undefined;
        }) => InfererArray | null;
    };
}
/**
 * Extended array object with typesafe infering and transform capability
 *
 * @export
 * @class InfererArray
 * @extends {Array}
 */
declare class InfererArray extends Array {
    private path;
    private value;
    private skipError;
    constructor(collection: DataType[], options: InferArrayOptions & {
        path: string;
        value: unknown;
    });
    /**
     * loop each item and parse or transform item using given function.
     *
     * @param {(row: Inferer, index: number, arr: InfererArray) => T} fn
     * @return {T[]} new transformed array
     */
    inferEach<T = DataType>(fn: (row: Inferer, index: number, arr: InfererArray) => T): T[];
    /**
     * extended Array.at that return inferer object.
     *
     * @param {number} index The zero-based index of the desired code unit. A negative index will count back from the last item.
     */
    inferAt(index: number): Inferer;
}

/**
 * Typesafe infer and transform data.
 */
declare function infer(data: unknown): Inferer;

declare class SchemaNode {
    /** Generated string scope */
    $scope: string;
    /** Description for the string scope */
    $description: string;
    /**
     * overriding json stringify
     *
     * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/JSON/stringify#tojson_behavior
     */
    toJSON(): unknown;
    constructor(scope: string, description: string);
}
declare class PermitNode extends SchemaNode {
    /** access control value for scope */
    private $value;
    /** hook function for subscribing to scope changes */
    private $subscriber;
    constructor(schema: SchemaNode, value?: boolean);
    /** helper for broadcasting changes to subscriber */
    private _broadcast;
    toJSON(): {
        scope: string;
        description: string;
        value: boolean;
    };
    /** readonly boolean value of true on disabled */
    get $disable(): boolean;
    /** boolean value of the scope. will broadcast changes when assigned new value to event listener. */
    get $able(): boolean;
    /** boolean value of the scope. will broadcast changes when assigned new value to event listener. */
    set $able(value: boolean);
    /** add listener function to scope that called on value changes */
    $subscribe(fn: (active: boolean, scope: SchemaScope) => void): () => boolean;
    /** utility for toggling value. will broadcast changes every call to event listener. */
    $toggle(): void;
}

declare class SchemaBranch<T extends GeneratedSchema> {
    readonly $schema: T;
    private _meta;
    constructor(schema: T, scope?: string, spacer?: string);
    toJSON(): unknown;
    get $meta(): {
        scope: string;
        spacer: string;
    };
    get $scope(): string;
    $each(nodeHook: (node: SchemaNode, key: string) => void, branchHook?: (branch: SchemaBranch<any>, key: string) => void): void;
    $refactor(prefix?: string, spacer?: string): void;
    $export(): {
        schema: SchemaExport<T>;
        meta: {
            scope: string;
            spacer: string;
        };
    };
    $model(): RevertSchema<T>;
    $permit(scopes?: (string | SchemaNode)[]): Permit<T>;
}

type SchemaScope = {
    $scope: string;
    $description: string;
};
/** Tree object for defining access control schema */
interface SchemaTree<T = string> extends Record<string, SchemaTree<T> | T> {
}
/** Generated tree schema type */
type GeneratedSchema<T extends SchemaTree = SchemaTree> = T extends SchemaTree ? {
    [P in keyof T]: T[P] extends SchemaTree ? Schema<GeneratedSchema<T[P]>> : SchemaNode;
} : never;
type RevertSchema<T> = T extends Record<string, any> ? {
    [K in OmitUtilityProps<keyof T>]: T[K] extends SchemaNode ? string : RevertSchema<T[K]>;
} : never;
type SchemaExport<T> = T extends Record<string, any> ? {
    [K in OmitUtilityProps<keyof T>]: T[K] extends SchemaNode ? {
        $scope: string;
        $description: string;
    } : SchemaExport<T[K]>;
} : never;
/** Access control schema tree */
type Schema<T extends GeneratedSchema> = T & SchemaBranch<T>;
type PermitNodes<T = Record<string, any>> = T extends Record<string, any> ? {
    [K in OmitUtilityProps<keyof T>]: T[K] extends SchemaNode ? PermitNode : T[K] extends {
        $scope: string;
        $description: string;
    } ? PermitNode : PermitNodes<T[K]> & PermitBranch<T[K]>;
} : never;
type PermitSchema<T extends Record<string, any>> = PermitNodes<T> & PermitBranch<T>;
/**
 * Infering literal string to scopes
 *
 * @template T
 */
type InferLiteralString<T, TD extends string> = T extends infer K ? K extends string ? K extends '' ? `${TD}` : `${K}` : `${TD}` : `${TD}`;
type ScopePrefixer<T, S> = T extends infer K ? K extends string ? K extends '' ? '' : `${K}${InferLiteralString<S, '.'>}` : never : never;
type UnionScopes<T, Prefix extends string = '', Spacer extends string = '.'> = T extends Record<string, unknown> ? {
    [K in keyof T]: K extends string ? T[K] extends Record<string, unknown> ? UnionScopes<T[K], `${ScopePrefixer<Prefix, Spacer>}${K}`, InferLiteralString<Spacer, '.'>> : `${ScopePrefixer<Prefix, Spacer>}${K}` : never;
}[keyof T] : Prefix;
type TransformSchema<T, V> = T extends SchemaTree ? {
    [P in keyof T]: T[P] extends SchemaTree ? TransformSchema<T[P], V> : V;
} : never;
type SchemaOptions = {
    /** add prefix to scope. default to '' (none) */
    prefix: string;
    /** string to use for spacer for scope. default to '.' */
    spacer: string;
};

declare class Permit<T extends SchemaTree<any>> {
    /** Access Schema */
    readonly on: PermitSchema<T>;
    private _nodes;
    private schema;
    private scopes;
    constructor(schema: T, scopes?: string[]);
    static getNodeScope(scope: string | SchemaNode | PermitNode): string;
    /** get nodes that has active access or granted */
    get granted(): PermitNode[];
    /** get nodes that is inactive or denied access */
    get denied(): PermitNode[];
    /** get list of permit node */
    get nodes(): PermitNode[];
    /** reset scope to default state */
    reset(): void;
    /** grant access to scope */
    grant(...scopes: (string | SchemaNode)[]): void;
    /** deny access to scope */
    deny(...scopes: (string | SchemaNode | PermitNode)[]): void;
    ableTo(scope: string | SchemaNode | PermitNode): boolean;
    get(scope: string | SchemaNode): PermitNode | undefined;
    has(scope: string | SchemaNode | PermitNode): boolean;
    child(scope: string): PermitNode[];
    /** loop through each node */
    $each(fn: (node: PermitNode) => void): void;
}
declare class PermitBranch<T extends Record<string, any>> {
    private $tree;
    constructor(nodes: PermitNodes<T>);
    $each(fn: (node: PermitNode) => void): void;
    get $able(): boolean;
    get $doable(): boolean;
    get $disable(): boolean;
    get $nodes(): PermitNode[];
    get $granted(): PermitNode[];
    get $denied(): PermitNode[];
}

/**
 * create access control schema from tree model. Return generated access control.
 *
 * ```js
 *
 * 	const schema = createSchema({
 * 		tweet: {
 * 			create: "create a blog post",
 * 			edit: "edit a blog post",
 * 			delete: "delete a blog post"
 * 		}
 * 	}, { prefix: 'com.tweeter', spacer: ':' })
 *
 * 	console.log(schema.tweet.create.$scope)
 * 	// output: 'com.tweeter:tweet:create'
 *
 *
 * ```
 */
declare function createSchema<T extends SchemaTree<string>>(model: T, options?: Partial<SchemaOptions>): Schema<GeneratedSchema<T>>;
/**
 * create permit object from exported schema.
 * usefull for ssr application when only need to create a permit on requested page.
 *
 * ```ts
 *
 * // on server side, send json data of schema access
 * const blogPageSchema = schema.blog.export().schema
 * const json = JSON.stringify(blogPageSchema)
 * type BlogPageSchema = typeof blogPageSchema
 *
 * // on client, parse json data
 * const schema = JSON.parse(json)
 * const permit = createPermit<BlogPageSchema>(schema, [])
 *
 *
 * ```
 */
declare function createPermit<T extends SchemaTree<SchemaScope>>(schema: T, scopes: string[]): Permit<T>;

type Ctx = Record<string, any>;
type ContextOptions = {
    /**
     * Option for shared context between procedure under the same context.
     * useful when context should be shared between procedure.
     *
     * When false, context will be isolated from each procedure and run when only needed.
     *
     * Default to true.
     */
    shared: boolean;
};
type ContextValue<V = any, T extends Ctx = Ctx> = T & {
    $value: V;
};
type ContextFunction<O, I extends any[], T extends Ctx> = (context: T, ...input: I) => O;
/** Procedure hook on after run called */
type OnAfterHook = (value: unknown) => void;
/** Procedure hook on before run called */
type OnBeforeHook<I extends any[], T extends Ctx> = (context: T, input: I) => void;
type ProcedureOptions<I extends any[], T extends Record<string, any>> = {
    onBeforeCalled: OnBeforeHook<I, T>;
    onAfterCalled: OnAfterHook;
};
/** Garda context used for injecting context to procedure */
declare class Context<T extends Ctx> {
    private _fn;
    private _context;
    private _updater;
    private shared;
    private procedureOpts;
    constructor(context: () => T, options?: Partial<ContextOptions>);
    /** context generator depending on context state */
    private _caller;
    /**
     * context function, alias for context()
     *
     * > ! notice. Different treatment on immidiate false/true.
     * >
     * > on immidiate option false (default), `context()` and `context.get()`
     * > will create a fresh context everytime. But when immidiate is set to true,
     * > context is already initiated. So it's not going to be isolated context.
     * */
    get(): T;
    /**
     * context utility for set context value
     *
     * > ! notice. Different tretment on immidiate false/true.
     * >
     * > on immidiate option true, when set is called. context isolated feature will be
     * > terminated because context is set to object that already initiated. And every update
     * > on context and it's object that context set to will be immidiately applied.
     * >
     * > Isolated feature of the context can be set again by calling `context.reset()`,
     * > and context will be set to initial context value.
     */
    set(context: T): void;
    /**
     * context utility for updating context value
     *
     * > ! notice. Different tretment on immidiate false/true.
     * >
     * > on immidiate option false (default). Given update will be treated as a hook function
     * > that only will be called when procedure is called as needed because of the isolated treatment.
     * >
     * > But when isolated feature is terminated by setting context through `context.set()`,
     * > update will be applied immidiately.
     */
    update(fn: (context: T) => void): void;
    /**
     * reset context to initial context initiation function.
     * this will remove all previous update to the context.
     * on immidiate set to true, context that is reseted will not point to previous initiated context.
     *
     * ```js
     * // Example
     * const context = createContext(() => {
     *  let count = 0
     *  return { setCount: (num) => count = num, count: () => count }
     * }, { immidiate: true })
     *
     * const ctx = context.get()
     *
     * ctx.setCount(10)
     * context().count() // this will return 10 too.
     *
     * context.update(c => c.setCount(20))
     * ctx.count() // this will return 20
     *
     * // on reset context
     * context.reset()
     * context().count() // will return 0
     * ctx.count() // will return 20
     *
     * ```
     */
    reset(): void;
    /** create procedure from context */
    procedure<O, I extends any[]>(fn: ContextFunction<O, I, T>): Procedure<O, I, T>;
    onAfterHook(fn: OnAfterHook): this;
    onBeforeHook(fn: OnBeforeHook<unknown[], T>): this;
}
declare class Procedure<O, I extends any[], T extends Ctx> {
    private _context;
    private _originalContext;
    private _rootFn;
    private _afterHookListener;
    private _beforeHookListener;
    private _runAfterListener;
    private _runBeforeListener;
    constructor(context: Context<T>, fn: ContextFunction<O, I, T>);
    /** Execute procedure function */
    get run(): (...input: I) => O;
    /**
     * Utility function for transforming and processing output from procedure.
     *
     * ---
     *
     * ```js
     *
     * // Usage
     * const getCount = context
     *  .procedure((ctx) => ctx.count)
     *
     * const doubleCount = getCount.pipe((value) => value * 2)
     * const isCountEven = getCount.pipe((value) => value % 2)
     *
     * const count = getCount() // return 1
     * const doubled = doubleCount() // return 2
     * const isEven = isCountEven() // return true
     *
     * ```
     */
    get pipe(): <V>(fn: (value: O) => V) => ((...input: I) => V) & {
        run: (...input: I) => V;
        pipe: <V_1>(fn: (value: V) => V_1) => ((...input: I) => V_1) & {
            run: (...input: I) => V_1;
            pipe: <V_2>(fn: (value: V_1) => V_2) => ((...input: I) => V_2) & {
                run: (...input: I) => V_2;
                pipe: <V_3>(fn: (value: V_2) => V_3) => ((...input: I) => V_3) & {
                    run: (...input: I) => V_3;
                    pipe: <V_4>(fn: (value: V_3) => V_4) => ((...input: I) => V_4) & {
                        run: (...input: I) => V_4;
                        pipe: <V_5>(fn: (value: V_4) => V_5) => ((...input: I) => V_5) & {
                            run: (...input: I) => V_5;
                            pipe: <V_6>(fn: (value: V_5) => V_6) => ((...input: I) => V_6) & {
                                run: (...input: I) => V_6;
                                pipe: <V_7>(fn: (value: V_6) => V_7) => ((...input: I) => V_7) & {
                                    run: (...input: I) => V_7;
                                    pipe: <V_8>(fn: (value: V_7) => V_8) => ((...input: I) => V_8) & {
                                        run: (...input: I) => V_8;
                                        pipe: <V_9>(fn: (value: V_8) => V_9) => ((...input: I) => V_9) & {
                                            run: (...input: I) => V_9;
                                            pipe: <V_10>(fn: (value: V_9) => V_10) => ((...input: I) => V_10) & {
                                                run: (...input: I) => V_10;
                                                pipe: <V_11>(fn: (value: V_10) => V_11) => ((...input: I) => V_11) & any;
                                            };
                                        };
                                    };
                                };
                            };
                        };
                    };
                };
            };
        };
    };
    /**
     * Utility for injecting updated context.
     *
     * useful when context is globally stored and updated,
     * but procedure need to be isolated.
     *
     * ---
     *
     * ```js
     *
     * // Example
     *
     * // isolated get data procedure with fetch
     * const getData = context
     *  .procedure((ctx) => fetch('/some-api', {
     *    headers: { "Authorization": `Bearer ${ctx.token}` }
     *  }))
     *
     * // get context and set token from cookie
     * const ctx = context()
     * ctx.token = cookie.get('Access-Token')
     * getData.inject(ctx) // inject updated context
     * const data = await getData()
     *
     * // or using context update utility
     * context.update((ctx) => {
     *  ctx.token = cookie.get('Access-Token)
     *  return ctx
     * })
     *
     * getData.inject(context)
     *
     * // or even better, directly update procedure context
     * getData.inject(ctx => {
     *  ctx.token = cookie.get('Access-Token)
     *  return ctx
     * })
     * ```
     */
    inject(context: T | Context<T> | ((context: T) => T)): void;
    /**
     * Utility for reseting context that has been injected
     * to original context used in current procedure.
     */
    resetContext(): void;
    onBeforeCalled(fn: OnBeforeHook<I, T>): () => void;
    onAfterCalled(fn: OnAfterHook): () => void;
    /**
     * Utility function for transforming and processing output from procedure.
     * Different than pipe utility, this utility give you context and input value.
     *
     * ---
     *
     * ```js
     *
     * // Usage
     * const getCount = sharedContext
     *  .procedure((ctx) => ctx.count)
     *
     * const doubleCount = getCount.use((ctx) => ctx.$value * 2)
     * const isCountEven = getCount.use((ctx) => ctx.count % 2)
     *
     * const count = getCount() // return 1
     * const doubled = doubleCount() // return 2
     * const isEven = isCountEven() // return true
     *
     * ```
     */
    use<V>(fn: ContextFunction<V, I, ContextValue<O, T>>): Procedure<V, I, T>;
    /**
     * Utility used for asyncronous return from procedure.
     * best use for data validation after fetching data using any data validation
     * like zod or any other data validation library.
     *
     * ---
     *
     * ```js
     *
     * // Example
     *
     * // isolated get data procedure with fetch
     * const getData = context
     *  .procedure((ctx) => fetch('/some-api', {
     *    headers: { "Authorization": `Bearer ${ctx.token}` }
     *  }))
     *  .useAsync(({ $value }) => {
     *    if (typeof $value !== 'object') return undefined
     *    return $value
     *  })
     *
     * // run getData procedure
     * const result = getData().then(value => {
     *  console.log(value) // -> value should be object or undefined
     * })
     * ```
     */
    useAsync<V>(fn: ContextFunction<V, I, ContextValue<Awaited<O>, T>>): Procedure<Promise<V>, I, T>;
    /**
     * Utility function for infering output and transform to new output type.
     *
     * ---
     *
     * ```ts
     *
     * // Usage
     * const getUserState = context
     *  .procedure((ctx, key: string) => ctx.getLocalStorageState(key))
     *  .infer(i => ({
     *    status: i.str('user.status.label'),
     *    hasWife: i.str('user.partner.gender') === 'female',
     *    hasChild: i.num('user.children.count') > 0
     *  }))
     *
     * const state = getUserState('some-key')
     *
     * ```
     */
    infer<V>(fn: (inferer: Inferer) => V): Procedure<V, I, T>;
    /**
     * Utility function for infering output and transform to new output type.
     * Equal to infer but for handling asyncronous output.
     *
     * ---
     *
     * ```ts
     *
     * // Usage
     * const getUserName = context
     *  .procedure(async (ctx, id: string) => await ctx.fetchUserData(id))
     *  .inferAsync(i =>
     *    i.str('data.fullName') + ' ' + i.str('data.lastName')
     *  )
     *
     * const name = await getUserName('some-user-id')
     *
     * ```
     */
    inferAsync<V>(fn: (inferer: Inferer) => V): Procedure<Promise<V>, I, T>;
}

/**
 * create garda procedure
 */
declare function createProcedure<O, I extends any[], T extends Record<string, any>>(context: (() => T) | Context<T>, fn: ContextFunction<O, I, T>, options?: Partial<ProcedureOptions<I, T>>): Procedure<O, I, T>;
/** Create garda context */
declare function createContext<T extends Record<string, any>>(context: () => T, options?: Partial<ContextOptions>): Context<T>;

export { Context, ContextFunction, ContextOptions, ContextValue, GeneratedSchema, InferLiteralString, Inferer, InfererArray, OnAfterHook, OnBeforeHook, Permit, PermitBranch, PermitNode, PermitNodes, PermitSchema, Procedure, RevertSchema, Schema, SchemaBranch, SchemaExport, SchemaNode, SchemaOptions, SchemaScope, SchemaTree, ScopePrefixer, TransformSchema, UnionScopes, createContext, createPermit, createProcedure, createSchema, infer, parseOptions };
