export interface AccessNode {
  readonly $scope: string;
  readonly $description: string;
}

export type SchemaModel = {
  [k: string]: SchemaModel | string | AccessNode;
};

export type AccessBranch<T extends SchemaModel> = {
  readonly $scope: string;
  readonly $model: SchemaModel;
  readonly $scopes: AccessNode[];
  $each: (fn: (node: AccessNode) => void) => void;
} & {
  [P in keyof T]: T[P] extends SchemaModel ? AccessBranch<T[P]> : AccessNode;
};

interface PermitNode extends AccessNode {
  readonly $ok: boolean;
  readonly $no: boolean;
}

type PermitBranch<T extends SchemaModel> = {
  readonly $some: boolean;
  readonly $none: boolean;
  readonly $all: boolean;
  readonly $scopes: AccessNode[];
} & {
  [K in keyof T]: T[K] extends SchemaModel ? PermitBranch<T[K]> : PermitNode;
};

type Permit<T extends SchemaModel> = {
  readonly granted: AccessNode[];
  readonly denied: AccessNode[];
  readonly scopes: AccessNode[];
  readonly is: PermitBranch<T>;

  reset: () => void;
  grant: (...scopes: string[]) => void;
  deny: (...scopes: string[]) => void;
  check: (scope: string) => boolean;
};

export function createPermit<T extends SchemaModel>(
  access: AccessBranch<T>,
  granted: string[] = []
): Permit<T> {
  const $value = new Map<string, boolean>();
  let permit: PermitBranch<T>;

  function generateNode(scope: string, description: string): PermitNode {
    return {
      get $scope() {
        return scope;
      },
      get $description() {
        return description;
      },
      get $ok() {
        return $value.get(scope) ?? false;
      },
      get $no() {
        return !$value.get(scope) ?? false;
      },
    };
  }

  function generate<N extends SchemaModel>(
    branch: AccessBranch<N>
  ): PermitBranch<N> {
    const childs: Record<string, PermitNode | PermitBranch<SchemaModel>> = {};
    for (const key in branch) {
      const cursor = branch[key];
      if (typeof cursor.$description === 'string') {
        const { $scope, $description } = cursor as AccessNode;
        childs[key] = generateNode($scope, $description);
      } else {
        childs[key] = generate(
          cursor as AccessBranch<SchemaModel>
        ) as PermitBranch<SchemaModel>;
      }
    }

    return {
      get $some() {
        return branch.$scopes.some((node) => {
          return $value.get(node.$scope) ?? false;
        });
      },
      get $none() {
        return branch.$scopes.every((node) => {
          if (!$value.has(node.$scope)) return false;
          return !$value.get(node.$scope) ?? false;
        });
      },
      get $all() {
        return branch.$scopes.every((node) => {
          return $value.get(node.$scope) ?? false;
        });
      },
      ...childs,
    } as PermitBranch<N>;
  }

  function mapAccessNode(value: boolean) {
    const scopes: string[] = [];
    for (const [key, val] of $value.entries()) {
      if (val === value) scopes.push(key);
    }

    return access.$scopes.filter((node) => scopes.includes(node.$scope));
  }

  function setScopeValue(value: boolean, ...scopes: string[]) {
    for (const scope of scopes) {
      if ($value.has(scope)) {
        $value.set(scope, value);
      }
    }

    // generate new permit every update
    permit = generate(access);
  }

  function reset() {
    $value.clear();
    access.$each((node) =>
      $value.set(node.$scope, granted.includes(node.$scope))
    );

    // generate new permit every reset
    permit = generate(access);
  }

  // First reset
  reset();

  return {
    get granted() {
      return mapAccessNode(true);
    },
    get denied() {
      return mapAccessNode(false);
    },
    get scopes() {
      return access.$scopes;
    },
    get is() {
      return permit;
    },
    reset() {
      reset();
      permit = generate(access);
    },
    grant(...scopes: string[]) {
      setScopeValue(true, ...scopes);
    },
    deny(...scopes: string[]) {
      setScopeValue(false, ...scopes);
    },
    check(scope: string) {
      return $value.get(scope) ?? false;
    },
  };
}

export function createSchema<T extends SchemaModel>(schema: T) {
  const spacer = '.';
  const prefix = '';

  function generateScope(...paths: string[]) {
    return paths.filter((p) => p && p !== '').join(spacer);
  }

  function generateAccess<N extends SchemaModel>(
    id: string,
    node: N,
    path = ''
  ): AccessBranch<N> {
    const $scope = generateScope(path, id);

    // generated childs for current branch node
    const childs: Record<string, AccessBranch<SchemaModel> | AccessNode> = {};
    for (const key in node) {
      const cursor: SchemaModel | string | AccessNode = node[key];
      if (typeof cursor === 'string') {
        childs[key] = {
          $scope: generateScope($scope, key),
          $description: cursor,
        } as AccessNode;
      } else if (
        typeof cursor.$scope === 'string' &&
        typeof cursor.$description === 'string'
      ) {
        childs[key] = {
          $scope: cursor.$scope,
          $description: cursor.$description,
        } as AccessNode;
      } else {
        childs[key] = generateAccess(
          key,
          cursor as SchemaModel,
          $scope
        ) as AccessBranch<SchemaModel>;
      }
    }

    function each(fn: (node: AccessNode) => void) {
      for (const key in childs) {
        const cursor = childs[key];
        if ((cursor as AccessBranch<SchemaModel>).$each !== undefined) {
          (cursor as AccessBranch<SchemaModel>).$each(fn);
        } else {
          fn(cursor as AccessNode);
        }
      }
    }

    const context = {
      get $scope() {
        return $scope;
      },
      get $scopes() {
        const scopes: AccessNode[] = [];
        each((node) => scopes.push(node));

        return scopes;
      },
      get $model() {
        // create duplicate schema model object
        const model = JSON.stringify(node);
        return JSON.parse(model) as N;
      },

      $each(fn: (node: AccessNode) => void) {
        each(fn);
      },

      ...childs,
    } as AccessBranch<N>;

    return context;
  }

  return generateAccess(prefix, schema);
}
